#!/usr/bin/env python

import os
import sys
import numpy as np
import argparse

##=============================================================================
class Tee(object):
    def __init__(self, *files):
        self.files = files
    def write(self, obj):
        for f in self.files:
            f.write(obj)
            f.flush() 
    def flush(self) :
        for f in self.files:
            f.flush()

f = open('observations.txt', 'w')
original = sys.stdout
sys.stdout = Tee(sys.stdout, f)
##=============================================================================


parser = argparse.ArgumentParser()
parser.add_argument('--fast', help='run without GUI',
                    action='store_true')
parser.add_argument('-nb', '--num_blockages', type=int, default=0)
parser.add_argument('-bls', '--block_sizes', type=str, default='250')
parser.add_argument('-blp', '--block_pos', type=str, default='750')
parser.add_argument('-bll', '--block_lanes', type=str, default='0')

args = parser.parse_args()

# file_path = 'observations.txt'
# sys.stdout = open(file_path, "w")


# """ Import SUMO library """
if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
    from traci.exceptions import TraCIException
    import traci.constants as tc
    if args.fast:
        print("Running without GUI")
        import libsumo as traci
    else:
        import traci as traci
else:
    sys.exit("please declare environment variable 'SUMO_HOME'")

sumo_cfg = './scenario_sublane/sumo.cfg.xml'#.format(pathlib.Path(__file__).parent.absolute())

sumo_parameters = ['sumo', '-c', sumo_cfg]

if args.fast:
    traci.start(sumo_parameters)
else:
    traci.init(9999)

### SIMULATION TIME SETTINGS ###
STEP = 0
START = 0.0
END = 86400.0

MEANING_OF_DIRECTIONS = {
    0 : 'stay',
    -1 : 'move right',
    +1 : 'move left'
}

### LANE CHANGING SETTINGS ###
LC_MODE = 512 # 1536 # 2560 # 1621 # 1541
LC_MAX_DURATION = 5 # END causes issues with perma deadlock???
NUMBER_OF_ACTIONS = 3 # stay on same lane, go to the left lane, go to the right lane

### SCENARIO SETTINGS ###
LENGTH_OF_ROAD = 1000
NUMBER_OF_VEH = 100
NUMBER_OF_LANES = 2

### AGENT SETTINGS ###
AGENT_PREFIX = 'agent'
class Agent(object):
    def __init__(self, id):
        self.ID = id
        self.actions = {
            0 : -1, # right
            1 : 0, # stay
            2 : +1, # left
        }
        self.prev_action = None        
        self.lane = None
        self.prev_lane = None
        self.target_lane = None

    def get_id(self):
        return self.ID


### LANE BLOCK SETTINGS ###
block_sizes = [float(x) for x in args.block_sizes.split(',')]
block_pos = [float(x) for x in args.block_pos.split(',')]
block_lanes = [int(x) for x in args.block_lanes.split(',')]
assert len(block_sizes) == len(block_pos) == len(block_lanes)

lane_block_settings = []
for i in range(len(block_sizes)):
    lane_block_settings.append( [block_sizes[i], block_pos[i], block_lanes[i]] )

if args.num_blockages > 0:
    for i, s in enumerate(lane_block_settings): # size, lane pos, lane index
        bl_size = s[0]
        bl_pos = s[1]
        bl_lane= s[2]

        type_id = 'ghost_{}'.format(i)
        traci.vehicletype.copy('DEFAULT_VEHTYPE', type_id)

        traci.vehicle.add(vehID='ghost{}'.format(i), routeID='r0', typeID='ghost', depart=0, departLane=bl_lane, departPos=bl_pos, departSpeed='random', arrivalLane='current', arrivalPos=bl_pos)

        traci.vehicle.setColor(typeID='ghost{}'.format(i),color=(255,0,0))
        traci.vehicle.setLength(typeID='ghost{}'.format(i),length=bl_size)
        traci.vehicle.setStop(vehID='ghost{}'.format(i), edgeID='1f2', pos=bl_pos, laneIndex=bl_lane, duration=10, until=END)


def isValidAction(agent, action):
    vh = agent.get_id()
    if action == 0:
        return (True, 'stay')
    else:
        vh = agent.get_id()
        lane_idx = traci.vehicle.getLaneIndex(vehID=vh)
        
        if 0 <= lane_idx + action < NUMBER_OF_LANES:
            lcs = traci.vehicle.getLaneChangeState(vehID=vh, direction=int(action))[1] # TraCI's lane change state for the given action
            lc_success = traci.vehicle.wantsAndCouldChangeLane(vehID=vh, direction=action, state=lcs) # Is it successful?
            if lc_success:
                return (True, 'can')
            else:
                return (False, 'cannot')
        else:
            return (False, 'invalid_lane')


vehicles_in_sim = []
vehicle_tagging = {}

def apply_action():

    for vh in traci.vehicle.getIDList():
        if 'ghost' in vh:
            continue

        if vh not in vehicles_in_sim:
            vehicles_in_sim.append(vh)
            agent = Agent(id=vh)
            vehicle_tagging[vh] = agent

        else:
            agent = vehicle_tagging[vh]

        l_pos = traci.vehicle.getLanePosition(vehID=vh)
        l_IDX = traci.vehicle.getLaneIndex(vehID=vh)

        agent.prev_lane = agent.lane
        agent.lane = l_IDX

        lateral_pos = traci.vehicle.getLateralLanePosition(vh)

        # if agent.target_lane == agent.lane:# and agent.prev_lane != agent.target_lane:
            # lateral_move = -lateral_pos
            # traci.vehicle.changeSublane(vh, lateral_move)
        
        action_choices = list(agent.actions.keys())
        action_IDX = np.random.choice(action_choices, p=None) # uniform sampling of actions
        # ended_on_target_lane_condition = (agent.target_lane == agent.lane and agent.target_lane != agent.prev_lane)
        # center_of_lane_condition = (lateral_pos == 0)

        traci.vehicle.changeLaneRelative(vehID=vh, indexOffset=agent.actions[action_IDX], duration=LC_MAX_DURATION)
        agent.target_lane = agent.lane + agent.actions[action_IDX]
        
        print('AGENT {}'.format(agent.get_id()))
        print('PREVIOUS LANE: {}'.format(agent.prev_lane))
        print('CURRENT LANE: {}'.format(agent.lane))
        print('TARGET LANE: {}'.format(agent.target_lane))
        print('ACTION SELECTED: {}'.format(MEANING_OF_DIRECTIONS[agent.actions[action_IDX]]))
        print('ACTION DIRECTION: {}'.format(agent.actions[action_IDX]))
        
        # valid_target_lane_condition = (0 <= agent.lane + agent.actions[action_IDX] < NUMBER_OF_LANES)
        # if valid_target_lane_condition:


        direction = int(agent.actions[action_IDX])
        laneChangeStateSumo, laneChangeStateTraCI = traci.vehicle.getLaneChangeState(vehID=vh, direction=direction)
        laneChangeStateSumo_pretty, laneChangeStateTraCI_pretty = traci.vehicle.getLaneChangeStatePretty(vehID=vh, direction=direction)

        couldChangeLane = traci.vehicle.couldChangeLane(vehID=vh, direction=direction, state=laneChangeStateTraCI)
        couldChangeLaneSumo = traci.vehicle.couldChangeLane(vehID=vh, direction=direction, state=laneChangeStateSumo)

        wantsAndCouldChangeLane = traci.vehicle.wantsAndCouldChangeLane(vehID=vh, direction=direction, state=laneChangeStateTraCI)
        wantsAndCouldChangeLaneSumo = traci.vehicle.wantsAndCouldChangeLane(vehID=vh, direction=direction, state=laneChangeStateSumo)


        print('LANE CHANGE STATE SUMO: {}'.format(laneChangeStateSumo_pretty))
        print('couldChangeLane SUMO: {}'.format(couldChangeLaneSumo))
        print('wantsAndCouldChangeLane SUMO: {}'.format(wantsAndCouldChangeLaneSumo))

        print('LANE CHANGE STATE TraCI: {}'.format(laneChangeStateTraCI_pretty))
        print('couldChangeLane TraCI: {}'.format(couldChangeLane))
        print('wantsAndCouldChangeLane TraCI: {}'.format(wantsAndCouldChangeLane))
        
        is_valid_action = isValidAction(agent, agent.actions[action_IDX])
        print('isValidAction {}'.format(is_valid_action))
        print('SPEED of vehicle {} : {}'.format(vh, traci.vehicle.getSpeed(vh)))
        print('\n')

def take_step(step):
    apply_action()
    
arrived = 0
while arrived < NUMBER_OF_VEH:

    print('################ SIM TIME: {} ##################'.format(traci.simulation.getTime()))
    print('################ SIM STEP COUNT: {} ################'.format(STEP))
    print('\n')

    for vh in traci.vehicle.getIDList():
        if 'ghost' not in vh:
            traci.vehicle.setLaneChangeMode(vh, LC_MODE)

    STEP+=1
    take_step(step=STEP)
    arrived += traci.simulation.getArrivedNumber()
    traci.simulationStep()
traci.close()


